from pathlib import Path
from typing import Optional
from xdg_base_dirs import xdg_config_home, xdg_data_home

# Store the custom root directory
_custom_root: Optional[Path] = None


def set_custom_root(path: Optional[str | Path]) -> None:
    """Set a custom root directory for all app files."""
    global _custom_root
    _custom_root = Path(path) if path else None


def _app_directory(root: Path) -> Path:
    """Create and return the application directory."""
    if _custom_root is not None:
        directory = _custom_root
    else:
        directory = root / "bagels"

    directory.mkdir(exist_ok=True, parents=True)
    return directory


def data_directory() -> Path:
    """Return (possibly creating) the application data directory."""
    return _app_directory(xdg_data_home())


def config_directory() -> Path:
    """Return (possibly creating) the application config directory."""
    return _app_directory(xdg_config_home())


def config_file() -> Path:
    return config_directory() / "config.yaml"


def database_file() -> Path:
    return data_directory() / "db.db"



from datetime import datetime, timedelta

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal
from textual.widgets import Label, Static

from textual.widgets import Button
from bagels.modals.input import InputModal
from bagels.config import CONFIG


class DateMode(Static):

    can_focus = True

    BINDINGS = [
        Binding(CONFIG.hotkeys.home.datemode.go_to_day, "go_to_day", "Go to Day"),
    ]

    FORM = [
        {
            "type": "dateAutoDay",
            "key": "date",
            "title": "Date",
            "default_value": datetime.now().strftime("%d"),
        }
    ]

    def __init__(self, parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="datemode-container", classes="module-container"
        )
        super().__setattr__("border_title", "Period")
        super().__setattr__(
            "border_subtitle", f"← {CONFIG.hotkeys.home.cycle_offset_type} →"
        )
        self.page_parent = parent
        self.first_day_of_week = CONFIG.defaults.first_day_of_week

    def on_mount(self) -> None:
        self.rebuild()

    # region Builder
    # -------------- Builder ------------- #

    # cell classes:
    # today, not_current_month, target, target_week, target_month

    def _get_month_days(self, date: datetime) -> list[tuple[datetime, bool]]:
        """Returns list of (date, is_current_month) tuples for calendar"""
        # Get first day of month
        first_day = date.replace(day=1)

        # Calculate days before month starts to fill calendar
        days_before = (first_day.weekday() - self.first_day_of_week) % 7
        start_date = first_day - timedelta(days=days_before)

        # Generate 42 days (6 weeks)
        days = []
        for i in range(42):
            curr_date = start_date + timedelta(days=i)
            is_current = curr_date.month == date.month
            days.append((curr_date, is_current))

        return days

    def rebuild(self) -> None:
        """Builds the calendar.

        The calendar should always give cells the following classes:
        * Date of today: "today"
        * Days not in the specified filter month: "not-current-month"

        Target date is a day that should have the class "target". The following are the rules:
        - Week: The cell's first day of the week (must respect first_day_of_week) if offset is not 0, else target date is today.
        - Month: First day of month if offset is not 0, else today.
        - Year: Always today

        The calendar should give cells the classes based on the current offset_type:
        - Week: The cell row's container of the current week: class "target_week"
        - Month: Every day of the specified filter month: "target_month"
        - Year: No rules.
        """
        filter_offset = self.page_parent.filter["offset"]
        filter_offset_type = self.page_parent.filter["offset_type"]

        today = datetime.now()
        target_date = self.page_parent.get_target_date()

        # Get calendar days
        calendar_days = self._get_month_days(target_date)

        # Update calendar labels
        calendar_rows = self.query(".calendar-row")

        for row_idx, row in enumerate(calendar_rows):
            row.remove_class("target_week")
            for col_idx, label in enumerate(row.query("Label")):
                day_idx = row_idx * 7 + col_idx
                if day_idx >= len(calendar_days):
                    continue

                date, is_current = calendar_days[day_idx]

                # Set day number
                label.update(str(date.day))

                # Reset classes
                label.set_classes("")

                # Add not current month class
                if not is_current:
                    label.add_class("not_current_month")

                # Add today class
                if date.date() == today.date():
                    label.add_class("today")

                # Add target class
                if date.date() == target_date.date():
                    label.add_class("target")

                # Add type-specific classes
                match filter_offset_type:
                    case "week":
                        # Calculate week start based on first_day_of_week
                        days_to_first = (
                            target_date.weekday() - self.first_day_of_week
                        ) % 7
                        week_start = target_date - timedelta(days=days_to_first)
                        week_end = week_start + timedelta(days=6)
                        if week_start.date() <= date.date() <= week_end.date():
                            row.add_class("target_week")
                    case "month":
                        if is_current:
                            label.add_class("target_month")

        self.page_parent.update_filter_label(self.query_one(".current-filter-label"))

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "prev-month":
            self.page_parent.action_dec_offset()
        elif event.button.id == "next-month":
            self.page_parent.action_inc_offset()
        self.rebuild()

    # region Callbacks
    # --------------- Callbacks --------------- #

    def action_go_to_day(self) -> None:
        def check_result(result) -> None:
            if result:
                self.page_parent.set_target_date(result["date"])

        self.app.push_screen(
            InputModal("Go to Day", form=self.FORM), callback=check_result
        )

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        with Container(classes="month-selector"):
            yield Button("<<<", id="prev-month")
            yield Label("Current Month", classes="current-filter-label")
            yield Button(">>>", id="next-month")

        with Static(classes="calendar"):
            # Create 6 rows of 7 days each for the calendar grid
            with Horizontal(classes="calendar-dotw-row"):
                days = ["M", "T", "W", "T", "F", "S", "S"]
                days = days[self.first_day_of_week :] + days[: self.first_day_of_week]
                for dotw in days:
                    yield Label(dotw)
            for _ in range(6):
                with Horizontal(classes="calendar-row"):
                    for _ in range(7):
                        yield Label("", classes="not_current_month")



import pytest
from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from bagels.models.database.db import Base
from bagels.models.account import Account
from bagels.models.record import Record
from bagels.models.split import Split
from bagels.models.person import Person
from bagels.models.category import Category, Nature
from bagels.managers import accounts

# Test fixtures
@pytest.fixture(scope="function")
def engine():
    """Create a test-specific database engine."""
    return create_engine("sqlite:///:memory:")

@pytest.fixture(scope="function")
def setup_test_database(engine):
    """Create all tables in the test database."""
    Base.metadata.create_all(engine)
    yield
    Base.metadata.drop_all(engine)

@pytest.fixture(scope="function")
def session(engine, setup_test_database):
    """Create a new session for a test."""
    Session = sessionmaker(bind=engine)
    session = Session()
    yield session
    session.close()

@pytest.fixture
def test_data(session):
    """Create test accounts, categories, and people."""
    # Create test accounts
    account1 = Account(name="Account 1", beginningBalance=1000.0)
    account2 = Account(name="Account 2", beginningBalance=500.0)
    session.add_all([account1, account2])
    
    # Create a test category
    category = Category(name="Test Category", nature=Nature.NEED, color="#FF0000")
    session.add(category)
    
    # Create a test person
    person = Person(name="Test Person")
    session.add(person)
    
    session.commit()
    
    return {
        "account1": account1,
        "account2": account2,
        "category": category,
        "person": person
    }

def test_basic_balance_calculation(session, test_data):
    """Test balance calculation with basic income and expense records."""
    # Create test records
    income_record = Record(
        label="Test Income",
        amount=200.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=True,
        date=datetime.now()
    )
    expense_record = Record(
        label="Test Expense",
        amount=150.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=False,
        date=datetime.now()
    )
    session.add_all([income_record, expense_record])
    session.commit()
    
    # Calculate expected balance: 1000 (beginning) + 200 (income) - 150 (expense) = 1050
    balance = accounts.get_account_balance(test_data["account1"].id, session)
    assert balance == 1050.0

def test_transfer_balance_calculation(session, test_data):
    """Test balance calculation with transfers between accounts."""
    # Create a transfer record
    transfer_record = Record(
        label="Test Transfer",
        amount=300.0,
        accountId=test_data["account1"].id,
        isTransfer=True,
        transferToAccountId=test_data["account2"].id,
        date=datetime.now()
    )
    session.add(transfer_record)
    session.commit()
    
    # Calculate expected balances
    # Account 1: 1000 (beginning) - 300 (transfer out) = 700
    balance1 = accounts.get_account_balance(test_data["account1"].id, session)
    assert balance1 == 700.0
    
    # Account 2: 500 (beginning) + 300 (transfer in) = 800
    balance2 = accounts.get_account_balance(test_data["account2"].id, session)
    assert balance2 == 800.0

def test_split_balance_calculation(session, test_data):
    """Test balance calculation with splits."""
    # Create a record with splits
    record = Record(
        label="Test Split Record",
        amount=300.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=False,
        date=datetime.now()
    )
    session.add(record)
    session.flush()
    
    # Create splits
    split1 = Split(
        recordId=record.id,
        amount=100.0,
        personId=test_data["person"].id,
        isPaid=True,
        accountId=test_data["account1"].id,
        paidDate=datetime.now()
    )
    split2 = Split(
        recordId=record.id,
        amount=100.0,
        personId=test_data["person"].id,
        isPaid=False,
        accountId=test_data["account1"].id
    )
    session.add_all([split1, split2])
    session.commit()
    
    # Calculate expected balance
    # 1000 (beginning) - 300 (expense) + 100 (paid split) = 800
    balance = accounts.get_account_balance(test_data["account1"].id, session)
    assert balance == 800.0

def test_combined_balance_calculation(session, test_data):
    """Test balance calculation with a combination of transactions."""
    # Create various transactions
    income_record = Record(
        label="Income",
        amount=500.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=True,
        date=datetime.now()
    )
    expense_record = Record(
        label="Expense",
        amount=200.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=False,
        date=datetime.now()
    )
    transfer_record = Record(
        label="Transfer",
        amount=300.0,
        accountId=test_data["account1"].id,
        isTransfer=True,
        transferToAccountId=test_data["account2"].id,
        date=datetime.now()
    )
    split_record = Record(
        label="Split Record",
        amount=400.0,
        accountId=test_data["account1"].id,
        categoryId=test_data["category"].id,
        isIncome=False,
        date=datetime.now()
    )
    session.add_all([income_record, expense_record, transfer_record, split_record])
    session.flush()
    
    # Create splits
    split = Split(
        recordId=split_record.id,
        amount=200.0,
        personId=test_data["person"].id,
        isPaid=True,
        accountId=test_data["account1"].id,
        paidDate=datetime.now()
    )
    session.add(split)
    session.commit()
    
    # Calculate expected balance for Account 1
    # 1000 (beginning)
    # + 500 (income)
    # - 200 (expense)
    # - 300 (transfer out)
    # - 400 (split record expense)
    # + 200 (paid split)
    # = 800
    balance1 = accounts.get_account_balance(test_data["account1"].id, session)
    assert balance1 == 800.0
    
    # Calculate expected balance for Account 2
    # 500 (beginning) + 300 (transfer in) = 800
    balance2 = accounts.get_account_balance(test_data["account2"].id, session)
    assert balance2 == 800.0



from datetime import datetime

from textual import events
from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import (
    Label,
    ListItem,
    ListView,
)

from bagels.components.fields import Fields
from bagels.managers.accounts import get_all_accounts_with_balance
from bagels.utils.validation import validateForm
from bagels.modals.base_widget import ModalContainer
from bagels.forms.form import Form, FormField


class Accounts(ListView):
    def __init__(
        self, accounts, initial_index: int = 0, type: str = "", *args, **kwargs
    ):
        super().__init__(
            *[
                ListItem(
                    Label(
                        str(account.name),
                        classes="name " + ("hidden-name" if account.hidden else ""),
                    ),
                    Label(
                        str(account.balance),
                        classes="balance "
                        + ("hidden-balance" if account.hidden else ""),
                    ),
                    classes="item",
                    id=f"account-{account.id}",
                )
                for account in accounts
            ],
            id=f"{type}-accounts",
            classes="accounts",
            initial_index=initial_index,
            *args,
            **kwargs,
        )


class TransferModal(ModalScreen):
    def __init__(self, record=None, *args, **kwargs):
        super().__init__(classes="modal-screen", *args, **kwargs)
        self.accounts = get_all_accounts_with_balance(get_hidden=True)
        self.form = Form(
            fields=[
                FormField(
                    title="Label",
                    key="label",
                    type="string",
                    placeholder="Label",
                    is_required=True,
                    default_value=str(record.label) if record else "",
                ),
                FormField(
                    title="Amount",
                    key="amount",
                    type="number",
                    placeholder="0.00",
                    min=0,
                    is_required=True,
                    default_value=str(record.amount) if record else "",
                ),
                FormField(
                    title="Date",
                    key="date",
                    type="dateAutoDay",
                    placeholder="dd (mm) (yy)",
                    default_value=(
                        record.date.strftime("%d")
                        if record
                        else datetime.now().strftime("%d")
                    ),
                ),
            ]
        )
        self.fromAccount = record.accountId if record else self.accounts[0].id
        self.toAccount = record.transferToAccountId if record else self.accounts[1].id
        if record:
            self.title = "Edit transfer"
        else:
            self.title = "New transfer"
        self.atAccountList = False

    def on_descendant_focus(self, event: events.DescendantFocus):
        id = event.widget.id
        if id.endswith("-accounts"):
            self.atAccountList = True
        else:
            self.atAccountList = False

    def on_key(self, event: events.Key):
        if self.atAccountList:
            if event.key == "right":
                self.screen.focus_next()
            elif event.key == "left":
                self.screen.focus_previous()
        else:
            if event.key == "up":
                self.screen.focus_previous()
            elif event.key == "down":
                self.screen.focus_next()
        if event.key == "enter":
            self.action_submit()
        elif event.key == "escape":
            self.dismiss(None)

    def on_list_view_highlighted(self, event: ListView.Highlighted):
        accountId = event.item.id.split("-")[1]
        if event.list_view.id == "from-accounts":
            self.fromAccount = accountId
        elif event.list_view.id == "to-accounts":
            self.toAccount = accountId

    def action_submit(self):
        resultForm, errors, isValid = validateForm(self, self.form)
        transfer_error_label = self.query_one("#transfer-error")
        # custom check for from and to accounts
        if self.fromAccount == self.toAccount:
            transfer_error_label.update("From and to accounts cannot be the same")
            transfer_error_label.add_class("active")
        else:
            transfer_error_label.update("")
            transfer_error_label.remove_class("active")
            if isValid:
                resultForm["accountId"] = self.fromAccount
                resultForm["transferToAccountId"] = self.toAccount
                resultForm["isTransfer"] = True
                self.dismiss(resultForm)
            else:
                previousErrors = self.query(".error")
                for error in previousErrors:
                    error.remove()
                for key, value in errors.items():
                    field = self.query_one(f"#row-field-{key}")
                    field.mount(Label(value, classes="error"))

    def compose(self) -> ComposeResult:
        yield ModalContainer(
            Container(
                Fields(self.form),
                Container(
                    Accounts(
                        self.accounts, initial_index=self.fromAccount, type="from"
                    ),
                    Label(">>>", classes="arrow"),
                    Accounts(self.accounts, initial_index=self.toAccount, type="to"),
                    classes="transfer-accounts-container",
                ),
                Label(id="transfer-error"),
                id="transfer-modal",
            ),
            custom_classes="wrapper max-width-80",
        )



COLORS = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_black",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
    "grey0",
    "gray0",
    "navy_blue",
    "dark_blue",
    "blue3",
    "blue1",
    "dark_green",
    "deep_sky_blue4",
    "dodger_blue3",
    "dodger_blue2",
    "green4",
    "spring_green4",
    "turquoise4",
    "deep_sky_blue3",
    "dodger_blue1",
    "green3",
    "spring_green3",
    "dark_cyan",
    "light_sea_green",
    "deep_sky_blue2",
    "deep_sky_blue1",
    "spring_green2",
    "cyan3",
    "dark_turquoise",
    "turquoise2",
    "green1",
    "spring_green1",
    "medium_spring_green",
    "cyan2",
    "cyan1",
    "dark_red",
    "deep_pink4",
    "purple4",
    "purple3",
    "blue_violet",
    "orange4",
    "grey37",
    "gray37",
    "medium_purple4",
    "slate_blue3",
    "royal_blue1",
    "chartreuse4",
    "dark_sea_green4",
    "pale_turquoise4",
    "steel_blue",
    "steel_blue3",
    "cornflower_blue",
    "chartreuse3",
    "cadet_blue",
    "sky_blue3",
    "steel_blue1",
    "pale_green3",
    "sea_green3",
    "aquamarine3",
    "medium_turquoise",
    "chartreuse2",
    "sea_green2",
    "sea_green1",
    "aquamarine1",
    "dark_slate_gray2",
    "dark_magenta",
    "dark_violet",
    "purple",
    "light_pink4",
    "plum4",
    "medium_purple3",
    "slate_blue1",
    "yellow4",
    "wheat4",
    "grey53",
    "gray53",
    "light_slate_grey",
    "light_slate_gray",
    "medium_purple",
    "light_slate_blue",
    "dark_olive_green3",
    "dark_sea_green",
    "light_sky_blue3",
    "sky_blue2",
    "dark_sea_green3",
    "dark_slate_gray3",
    "sky_blue1",
    "chartreuse1",
    "light_green",
    "pale_green1",
    "dark_slate_gray1",
    "red3",
    "medium_violet_red",
    "magenta3",
    "dark_orange3",
    "indian_red",
    "hot_pink3",
    "medium_orchid3",
    "medium_orchid",
    "medium_purple2",
    "dark_goldenrod",
    "light_salmon3",
    "rosy_brown",
    "grey63",
    "gray63",
    "medium_purple1",
    "gold3",
    "dark_khaki",
    "navajo_white3",
    "grey69",
    "gray69",
    "light_steel_blue3",
    "light_steel_blue",
    "yellow3",
    "dark_sea_green2",
    "light_cyan3",
    "light_sky_blue1",
    "green_yellow",
    "dark_olive_green2",
    "dark_sea_green1",
    "pale_turquoise1",
    "deep_pink3",
    "magenta2",
    "hot_pink2",
    "orchid",
    "medium_orchid1",
    "orange3",
    "light_pink3",
    "pink3",
    "plum3",
    "violet",
    "light_goldenrod3",
    "tan",
    "misty_rose3",
    "thistle3",
    "plum2",
    "khaki3",
    "light_goldenrod2",
    "light_yellow3",
    "grey84",
    "gray84",
    "light_steel_blue1",
    "yellow2",
    "dark_olive_green1",
    "honeydew2",
    "light_cyan1",
    "red1",
    "deep_pink2",
    "deep_pink1",
    "magenta1",
    "orange_red1",
    "indian_red1",
    "hot_pink",
    "dark_orange",
    "salmon1",
    "light_coral",
    "pale_violet_red1",
    "orchid2",
    "orchid1",
    "orange1",
    "sandy_brown",
    "light_salmon1",
    "light_pink1",
    "pink1",
    "plum1",
    "gold1",
    "navajo_white1",
    "misty_rose1",
    "thistle1",
    "yellow1",
    "light_goldenrod1",
    "khaki1",
    "wheat1",
    "cornsilk1",
    # "grey100",
    # "gray100",
    # "grey3",
    # "gray3",
    # "grey7",
    # "gray7",
    # "grey11",
    # "gray11",
    # "grey15",
    # "gray15",
    # "grey19",
    # "gray19",
    # "grey23",
    # "gray23",
    # "grey27",
    # "gray27",
    # "grey30",
    # "gray30",
    # "grey35",
    # "gray35",
    # "grey39",
    # "gray39",
    # "grey42",
    # "gray42",
    # "grey46",
    # "gray46",
    # "grey50",
    # "gray50",
    # "grey54",
    # "gray54",
    # "grey58",
    # "gray58",
    # "grey62",
    # "gray62",
    # "grey66",
    # "gray66",
    # "grey70",
    # "gray70",
    # "grey74",
    # "gray74",
    # "grey78",
    # "gray78",
    # "grey82",
    # "gray82",
    # "grey85",
    # "gray85",
    # "grey89",
    # "gray89",
    # "grey93",
    # "gray93",
]



from pathlib import Path
from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Label, MarkdownViewer, Rule, Static
from bagels.bagel import get_string, pprint, render_frame, theta_spacing, phi_spacing


class Welcome(Static):

    A = B = 1
    can_focus = True

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, id="welcome-container")
        file_path = Path(__file__).parent.parent.parent / "static" / "welcome.md"
        with open(file_path, "r") as file:
            self.welcome_text = file.read()

    def on_mount(self) -> None:
        self.set_interval(1 / 10, self.update_bagel)

    def update_bagel(self) -> None:
        bagel = self.query_one("#bagel")
        self.A += theta_spacing
        self.B += phi_spacing
        bagel.update(get_string(render_frame(self.A, 1)))

    def compose(self) -> ComposeResult:

        with Container(classes="text-container"):
            yield MarkdownViewer(self.welcome_text, show_table_of_contents=False)
        with Container(classes="bagel-container"):
            yield Label(id="bagel")



from datetime import datetime
from sqlalchemy import and_, select
from sqlalchemy.orm import joinedload, sessionmaker

from bagels.models.database.app import db_engine
from bagels.models.person import Person
from bagels.models.record import Record
from bagels.models.split import Split
from bagels.managers.utils import get_start_end_of_period

Session = sessionmaker(bind=db_engine)


def create_person(data):
    """Create a new person entry in the database."""
    session = Session()
    try:
        new_person = Person(**data)
        session.add(new_person)
        session.commit()
        session.refresh(new_person)
        session.expunge(new_person)
        return new_person
    finally:
        session.close()


def get_all_persons():
    """Retrieve all persons from the database."""
    session = Session()
    try:
        return session.scalars(select(Person)).all()
    finally:
        session.close()


def get_person_by_id(person_id):
    """Retrieve a person by their ID."""
    session = Session()
    try:
        return session.get(Person, person_id)
    finally:
        session.close()


def update_person(person_id, data):
    """Update a person's information by their ID."""
    session = Session()
    try:
        person = session.get(Person, person_id)
        if person:
            for key, value in data.items():
                setattr(person, key, value)
            session.commit()
            session.refresh(person)
            session.expunge(person)
        return person
    finally:
        session.close()


def delete_person(person_id):
    """Delete a person from the database by their ID."""
    session = Session()
    try:
        person = session.get(Person, person_id)
        if person:
            session.delete(person)
            session.commit()
            return True
        return False
    finally:
        session.close()


def get_persons_with_splits(offset: int = 0, offset_type: str = "month"):
    """Get all persons with their splits for the specified period."""
    session = Session()
    try:
        start_of_period, end_of_period = get_start_end_of_period(offset, offset_type)
        result = (
            session.scalars(
                select(Person)
                .options(
                    joinedload(Person.splits)
                    .joinedload(Split.record)
                    .joinedload(Record.category),
                    joinedload(Person.splits).joinedload(Split.account),
                )
                .join(Person.splits)
                .join(Split.record)
                .filter(and_(Record.date >= start_of_period, Record.date < end_of_period))
                .order_by(Record.date.asc())
                .distinct()
            )
        )
        return result.unique().all()
    finally:
        session.close()



import numpy as np

screen_size = 15
theta_spacing = 0.07  # Spacing for theta, the angle for the circle around R1
phi_spacing = 0.02  # Spacing for phi, the angle for the rotation around R2
illumination = np.fromiter(".,-~:;=!$░▒▓", dtype="<U1")

A = 1
B = 1
R1 = 1
R2 = 2
K2 = 5
K1 = screen_size * K2 * 3 / (8 * (R1 + R2))


def render_frame(A: float, B: float) -> np.ndarray:
    """
    Returns a frame of the spinning 3D donut.
    Based on the pseudocode from: https://www.a1k0n.net/2011/07/20/donut-math.html
    """
    cos_A = np.cos(A)
    sin_A = np.sin(A)
    cos_B = np.cos(B)
    sin_B = np.sin(B)

    output = np.full((screen_size, screen_size), " ")  # (40, 40)
    zbuffer = np.zeros((screen_size, screen_size))  # (40, 40)

    cos_phi = np.cos(phi := np.arange(0, 2 * np.pi, phi_spacing))  # (315,)
    sin_phi = np.sin(phi)  # (315,)
    cos_theta = np.cos(theta := np.arange(0, 2 * np.pi, theta_spacing))  # (90,)
    sin_theta = np.sin(theta)  # (90,)
    circle_x = R2 + R1 * cos_theta  # (90,)
    circle_y = R1 * sin_theta  # (90,)

    x = (
        np.outer(cos_B * cos_phi + sin_A * sin_B * sin_phi, circle_x)
        - circle_y * cos_A * sin_B
    ).T  # (90, 315)
    y = (
        np.outer(sin_B * cos_phi - sin_A * cos_B * sin_phi, circle_x)
        + circle_y * cos_A * cos_B
    ).T  # (90, 315)
    z = ((K2 + cos_A * np.outer(sin_phi, circle_x)) + circle_y * sin_A).T  # (90, 315)
    ooz = np.reciprocal(z)  # Calculates 1/z
    xp = (screen_size / 2 + K1 * ooz * x).astype(int)  # (90, 315)
    yp = (screen_size / 2 - K1 * ooz * y).astype(int)  # (90, 315)
    L1 = (
        (np.outer(cos_phi, cos_theta) * sin_B) - cos_A * np.outer(sin_phi, cos_theta)
    ) - sin_A * sin_theta  # (315, 90)
    L2 = cos_B * (cos_A * sin_theta - np.outer(sin_phi, cos_theta * sin_A))  # (315, 90)
    L = np.around(((L1 + L2) * 8)).astype(int).T  # (90, 315)
    mask_L = L >= 0  # (90, 315)
    chars = illumination[L]  # (90, 315)

    for i in range(90):
        mask = mask_L[i] & (ooz[i] > zbuffer[xp[i], yp[i]])  # (315,)

        zbuffer[xp[i], yp[i]] = np.where(mask, ooz[i], zbuffer[xp[i], yp[i]])
        output[xp[i], yp[i]] = np.where(mask, chars[i], output[xp[i], yp[i]])

    return output


def pprint(array: np.ndarray) -> None:
    """Pretty print the frame."""
    print(*[" ".join(row) for row in array], sep="\n")


def get_string(array: np.ndarray) -> str:
    """Return the frame as a string."""
    return "\n".join([" ".join(row) for row in array])


if __name__ == "__main__":
    for _ in range(screen_size * screen_size):
        A += theta_spacing
        B += phi_spacing
        print("\x1b[H")
        pprint(render_frame(A, B))



from datetime import datetime, timedelta

from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.widgets import Label, Static

from bagels.components.barchart import Barchart, BarchartData
from bagels.components.percentage_bar import PercentageBar, PercentageBarItem
from bagels.config import CONFIG
from bagels.managers.categories import get_all_categories_records
from bagels.managers.utils import (
    get_period_average,
    get_period_figures,
    get_start_end_of_period,
)


class Insights(Static):

    BINDINGS = [
        (
            CONFIG.hotkeys.home.insights.toggle_use_account,
            "toggle_use_account",
            "Toggle use account",
        )
    ]

    can_focus = True

    def __init__(self, parent: Static, *args, **kwargs) -> None:
        super().__init__(
            *args, **kwargs, id="insights-container", classes="module-container"
        )
        super().__setattr__("border_title", "Insights")
        self.page_parent = parent
        self.use_account = False  # insights of specific account if True

    def on_mount(self) -> None:
        self.rebuild()

    # region Builder
    # -------------- Builder ------------- #

    def rebuild(self) -> None:
        period_net = self._update_labels()
        items = self.get_percentage_bar_items(period_net)
        self.percentage_bar.set_items(items)
        data = self.get_period_barchart_data()
        self.period_barchart.set_data(data)

    def _update_labels(self) -> None:
        current_filter_label = self.query_one(".current-filter-label")
        period_net_label = self.query_one(".period-net")
        period_average_label = self.query_one(".period-average")
        average_label = self.query_one(".average-label")

        mode_isIncome = self.page_parent.mode["isIncome"]
        label = "Income" if mode_isIncome else "Expense"

        if self.use_account:
            current_filter_label.update(
                f"{self.page_parent.mode['accountId']['default_value_text']} {label} of {self.page_parent.get_filter_label()}"
            )
        else:
            current_filter_label.update(
                f"{label} of {self.page_parent.get_filter_label()}"
            )
        average_label.update(f"{label} per day")

        if self.use_account:
            params = {
                **self.page_parent.filter,
                "accountId": self.page_parent.mode["accountId"]["default_value"],
                "isIncome": mode_isIncome,
            }
        else:
            params = {**self.page_parent.filter, "isIncome": mode_isIncome}

        period_net = get_period_figures(**params)
        period_average = get_period_average(period_net, **self.page_parent.filter)
        period_net_label.update(str(period_net))
        period_average_label.update(str(period_average))

        return period_net

    def get_percentage_bar_items(
        self, period_net=1, limit: int = 5
    ) -> list[PercentageBarItem]:
        if period_net == 0:
            return []
        if self.use_account:
            category_records = get_all_categories_records(
                **self.page_parent.filter,
                is_income=self.page_parent.mode["isIncome"],
                account_id=self.page_parent.mode["accountId"]["default_value"],
            )
        else:
            category_records = get_all_categories_records(
                **self.page_parent.filter,
                is_income=self.page_parent.mode["isIncome"],
            )

        # Sort categories by percentage in descending order
        items = []
        if len(category_records) <= limit:
            # If we have 5 or fewer categories, show them all
            for category in category_records:
                items.append(
                    PercentageBarItem(
                        name=category.name,
                        count=int(category.amount / period_net * 100),
                        color=category.color,
                    )
                )
        else:
            # Show top 4 categories and group the rest as "Others"
            for category in category_records[:limit]:
                items.append(
                    PercentageBarItem(
                        name=category.name,
                        count=int(category.amount / period_net * 100),
                        color=category.color,
                    )
                )

            # Sum up the amounts of remaining categories
            others_amount = sum(cat.amount for cat in category_records[limit:])
            others_percentage = int(others_amount / period_net * 100)
            items.append(
                PercentageBarItem(name="Others", count=others_percentage, color="white")
            )

        return items

    def get_period_barchart_data(self) -> BarchartData:
        offset_type = self.page_parent.filter["offset_type"]
        offset = self.page_parent.filter["offset"]
        if offset_type == "day":
            return BarchartData(amounts=[], labels=[])

        # Get data for each sub-period
        amounts = []
        labels = []

        match offset_type:
            case "year":
                # Get start of the target year
                start_date = datetime.now().replace(
                    month=1, day=1, year=datetime.now().year + offset
                )
                for i in range(12):
                    # Calculate month offset relative to today
                    target_date = start_date.replace(month=i + 1)
                    month_offset = (target_date.year - datetime.now().year) * 12 + (
                        target_date.month - datetime.now().month
                    )

                    amount = get_period_figures(
                        offset_type="month",
                        offset=month_offset,
                        isIncome=self.page_parent.mode["isIncome"],
                    )
                    amounts.append(abs(amount))
                    labels.append(target_date.strftime("%b"))
            case "week":
                start_date, end_date = get_start_end_of_period(offset, offset_type)

                for i in range(7):
                    current_date = start_date + timedelta(days=i)
                    day_offset = (current_date - datetime.now()).days + 1

                    amount = get_period_figures(
                        offset_type="day",
                        offset=day_offset,
                        isIncome=self.page_parent.mode["isIncome"],
                    )
                    amounts.append(abs(amount))
                    labels.append(current_date.strftime("%d"))

        return BarchartData(amounts=amounts, labels=labels)

    # region Callbacks
    # ------------- callbacks ------------ #
    def action_toggle_use_account(self) -> None:
        self.use_account = not self.use_account
        self.rebuild()

    # region View
    # --------------- View --------------- #

    def compose(self) -> ComposeResult:
        with Horizontal(classes="figures-container"):
            with Container(classes="net container"):
                yield Label(classes="current-filter-label title")  # dynamic
                yield Label("Loading...", classes="period-net amount")  # dynamic
            with Container(classes="average container"):
                yield Label("<> per day", classes="average-label title")  # dynamic
                yield Label("Loading...", classes="period-average amount")  # dynamic

        self.percentage_bar = PercentageBar()
        self.period_barchart = Barchart()
        yield self.percentage_bar
        yield self.period_barchart



import warnings
from typing import Any, Literal
import yaml
from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic_settings_yaml import YamlBaseSettings
from bagels.locations import config_file


class Defaults(BaseModel):
    period: Literal["day", "week", "month", "year"] = "week"
    first_day_of_week: int = Field(ge=0, le=6, default=6)
    date_format: str = "%d/%m"


class InsightsHotkeys(BaseModel):
    toggle_use_account: str = "\\"


class DatemodeHotkeys(BaseModel):
    go_to_day: str = "g"


class HomeHotkeys(BaseModel):
    categories: str = "c"
    new_transfer: str = "t"
    toggle_splits: str = "s"
    display_by_date: str = "q"
    display_by_person: str = "w"
    cycle_offset_type: str = "."
    toggle_income_mode: str = "/"
    select_prev_account: str = "["
    select_next_account: str = "]"
    insights: InsightsHotkeys = InsightsHotkeys()
    datemode: DatemodeHotkeys = DatemodeHotkeys()


class RecordModalHotkeys(BaseModel):
    new_split: str = "ctrl+a"
    new_paid_split: str = "ctrl+s"
    delete_last_split: str = "ctrl+d"


class CategoriesHotkeys(BaseModel):
    new_subcategory: str = "s"
    browse_defaults: str = "b"


class Hotkeys(BaseModel):
    new: str = "a"
    delete: str = "d"
    edit: str = "e"
    toggle_jump_mode: str = "v"
    home: HomeHotkeys = HomeHotkeys()
    record_modal: RecordModalHotkeys = RecordModalHotkeys()
    categories: CategoriesHotkeys = CategoriesHotkeys()


class Symbols(BaseModel):
    line_char: str = "│"
    finish_line_char: str = "╰"
    split_paid: str = "✓"
    split_unpaid: str = "⨯"
    category_color: str = "●"
    amount_positive: str = "+"
    amount_negative: str = "-"


class State(BaseModel):
    theme: str = "posting"


class Config(YamlBaseSettings):
    hotkeys: Hotkeys = Hotkeys()
    symbols: Symbols = Symbols()
    defaults: Defaults = Defaults()
    state: State = State()
    model_config = SettingsConfigDict(
        yaml_file=str(config_file()),
        yaml_file_encoding="utf-8",
    )

    def __init__(self, **data):
        super().__init__(**data)
        self.ensure_yaml_fields()

    def ensure_yaml_fields(self):
        # Load current config or create a new one if it doesn't exist
        try:
            with open(config_file(), "r") as f:
                config = yaml.safe_load(f) or {}
        except FileNotFoundError:
            config = {}

        # Update config with default values for missing fields
        def update_config(default, current):
            for key, value in default.items():
                if isinstance(value, dict):
                    current[key] = update_config(value, current.get(key, {}))
                elif key not in current:
                    current[key] = value
            return current

        default_config = self.model_dump()
        config = update_config(default_config, config)

        # Write back to the YAML file
        with open(config_file(), "w") as f:
            yaml.dump(config, f, default_flow_style=False)

    @classmethod
    def get_default(cls):
        # Create a default instance without reading from file
        return cls.model_construct(
            hotkeys=Hotkeys(), symbols=Symbols(), defaults=Defaults(), state=State()
        )


# Only try to load from file if it exists
config_path = config_file()

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    CONFIG = (
        Config.get_default() if not config_path.exists() else Config()
    )  # ignore warnings about empty env file


def write_state(key: str, value: Any) -> None:
    """Write a state value to the config.yaml file."""
    try:
        with open(config_file(), "r") as f:
            config = yaml.safe_load(f) or {}
    except FileNotFoundError:
        config = {}

    if "state" not in config:
        config["state"] = {}
    config["state"][key] = value

    with open(config_file(), "w") as f:
        yaml.dump(config, f, default_flow_style=False)



# file is used with textual run --dev

from pathlib import Path
from bagels.locations import set_custom_root

if __name__ == "__main__":
    # set_custom_root(Path("./instance/"))

    from bagels.models.database.app import init_db

    init_db()

    from bagels.app import App

    app = App()
    app.run()



import copy
from datetime import datetime

from rich.text import Text

from bagels.constants import COLORS
from bagels.models.category import Nature
from bagels.managers.categories import get_category_by_id
from bagels.forms.form import Form, FormField, Option, Options


class CategoryForm:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    # ------------ Blueprints ------------ #

    FORM = Form(
        fields=[
            FormField(
                placeholder="My Category",
                title="Name",
                key="name",
                type="string",
                is_required=True,
            ),
            FormField(
                title="Nature",
                key="nature",
                type="autocomplete",
                options=Options(
                    items=[
                        Option(
                            text="Must",
                            value=Nature.MUST,
                            prefix=Text("●", style="red"),
                        ),
                        Option(
                            text="Need",
                            value=Nature.NEED,
                            prefix=Text("●", style="orange"),
                        ),
                        Option(
                            text="Want",
                            value=Nature.WANT,
                            prefix=Text("●", style="green"),
                        ),
                    ]
                ),
                is_required=True,
                placeholder="Select Nature",
            ),
            FormField(
                title="Color",
                key="color",
                type="autocomplete",
                options=Options(
                    items=[
                        Option(value=color, prefix=Text("●", style=color))
                        for color in COLORS
                    ]
                ),
                is_required=True,
                placeholder="Select Color",
            ),
        ]
    )

    # ----------------- - ---------------- #

    # def __init__(self):
    # self._populate_form_options()

    # -------------- Helpers ------------- #

    # ------------- Builders ------------- #

    def get_subcategory_form(self, parent_id: int) -> Form:
        subcategory_form = copy.deepcopy(self.FORM)
        subcategory_form.fields.append(
            FormField(
                key="parentCategoryId",
                type="hidden",
                default_value=str(parent_id),
            )
        )
        return subcategory_form

    def get_filled_form(self, category_id: int) -> Form:
        """Return a copy of the form with values from the record"""
        filled_form = copy.deepcopy(self.FORM)
        category = get_category_by_id(category_id)
        if category:
            for field in filled_form.fields:
                value = getattr(category, field.key)
                if field.key == "nature":
                    field.default_value = category.nature
                    field.default_value_text = category.nature.value
                elif field.key == "color":
                    field.default_value = category.color
                    field.default_value_text = category.color
                else:
                    field.default_value = str(value) if value is not None else ""

        return filled_form

    def get_form(self):
        """Return the base form with default values"""
        return self.FORM



from importlib.metadata import metadata

from rich.console import Group
from rich.text import Text
from textual import events, log, on
from textual.app import App as TextualApp
from textual.app import ComposeResult
from textual.command import CommandPalette
from textual.containers import Container
from textual.content import Content
from textual.css.query import NoMatches
from textual.geometry import Size
from textual.reactive import Reactive, reactive
from textual.signal import Signal
from textual.widget import Widget
from textual.widgets import Footer, Label, Tabs

from bagels.components.jump_overlay import JumpOverlay
from bagels.components.jumper import Jumper
from bagels.config import CONFIG, write_state
from bagels.home import Home
from bagels.modals.categories import CategoriesModal
from bagels.provider import AppProvider
from bagels.themes import BUILTIN_THEMES, Theme
from bagels.utils.user_host import get_user_host_string


class App(TextualApp):

    CSS_PATH = "index.tcss"
    BINDINGS = [
        (CONFIG.hotkeys.toggle_jump_mode, "toggle_jump_mode", "Jump Mode"),
        (CONFIG.hotkeys.home.categories, "go_to_categories", "Categories"),
        ("ctrl+q", "quit", "Quit"),
    ]
    COMMANDS = {AppProvider}

    app_theme: Reactive[str] = reactive(CONFIG.state.theme, init=False)
    """The currently selected theme. Changing this reactive should
    trigger a complete refresh via the `watch_theme` method."""
    layout: Reactive[str] = reactive("h")
    """The current layout of the app."""
    _jumping: Reactive[bool] = reactive(False, init=False, bindings=True)
    """True if 'jump mode' is currently active, otherwise False."""

    # region init
    def __init__(self):
        # Initialize available themes with a default
        available_themes: dict[str, Theme] = {"galaxy": BUILTIN_THEMES["galaxy"]}
        available_themes |= BUILTIN_THEMES
        self.themes = available_themes
        super().__init__()

        # Get package metadata directly
        meta = metadata("bagels")
        self.project_info = {"name": meta["Name"], "version": meta["Version"]}

    def on_mount(self) -> None:
        # --------------- theme -------------- #
        self.theme_change_signal = Signal[Theme](self, "theme-changed")
        # -------------- jumper -------------- #
        self.jumper = Jumper(
            {
                "accounts-container": "a",
                "insights-container": "i",
                "records-container": "r",
                "templates-container": "t",
                # "incomemode-container": "v",
                "datemode-container": "p",
            },
            screen=self.screen,
        )

    # used by the textual app to get the theme variables
    def get_css_variables(self) -> dict[str, str]:
        if self.app_theme:
            theme = self.themes.get(self.app_theme)
            if theme:
                color_system = theme.to_color_system().generate()
            else:
                color_system = {}
        else:
            color_system = {}
        return {**super().get_css_variables(), **color_system}

    def command_theme(self, theme: str) -> None:
        self.app_theme = theme
        self.notify(
            f"Theme is now [b]{theme!r}[/].", title="Theme updated", timeout=2.5
        )
        write_state("theme", theme)

    # region theme
    # --------------- theme -------------- #
    def watch_app_theme(self, theme: str | None) -> None:
        self.refresh_css(animate=False)
        self.screen._update_styles()
        if theme:
            theme_object = self.themes[theme]
            self.theme_change_signal.publish(theme_object)

    @on(CommandPalette.Opened)
    def palette_opened(self) -> None:
        self._original_theme = self.app_theme

    @on(CommandPalette.OptionHighlighted)
    def palette_option_highlighted(
        self, event: CommandPalette.OptionHighlighted
    ) -> None:
        prompt: Content = event.highlighted_event.option.prompt
        command_name = prompt.plain.split("\n")[0]
        if ":" in command_name:
            name, value = command_name.split(":", maxsplit=1)
            name = name.strip()
            value = value.strip()
            if name == "theme":
                if value in self.themes:
                    self.app_theme = value
            else:
                self.app_theme = self._original_theme

    @on(CommandPalette.Closed)
    def palette_closed(self, event: CommandPalette.Closed) -> None:
        # If we closed with a result, that will be handled by the command
        # being triggered. However, if we closed the palette with no result
        # then make sure we revert the theme back.
        # if not self.settings.command_palette.theme_preview:
        #     return
        if not event.option_selected:
            self.app_theme = self._original_theme

    # region jumper
    # -------------- jumper -------------- #
    def action_toggle_jump_mode(self) -> None:
        self._jumping = not self._jumping

    def watch__jumping(self, jumping: bool) -> None:
        focused_before = self.focused
        if focused_before is not None:
            self.set_focus(None, scroll_visible=False)

        def handle_jump_target(target: str | Widget | None) -> None:
            if isinstance(target, str):
                try:
                    target_widget = self.screen.query_one(f"#{target}")
                except NoMatches:
                    log.warning(
                        f"Attempted to jump to target #{target}, but it couldn't be found on {self.screen!r}"
                    )
                else:
                    if target_widget.focusable:
                        self.set_focus(target_widget)
                    else:
                        target_widget.post_message(
                            events.Click(0, 0, 0, 0, 0, False, False, False)
                        )

            elif isinstance(target, Widget):
                self.set_focus(target)
            else:
                # If there's no target (i.e. the user pressed ESC to dismiss)
                # then re-focus the widget that was focused before we opened
                # the jumper.
                if focused_before is not None:
                    self.set_focus(focused_before, scroll_visible=False)

        self.clear_notifications()
        self.push_screen(JumpOverlay(self.jumper), callback=handle_jump_target)

    # region hooks
    # --------------- hooks -------------- #

    def on_tabs_tab_activated(self, event: Tabs.TabActivated) -> None:
        if event.tab.id.startswith("t"):
            activeIndex = int(event.tab.id.removeprefix("t")) - 1
            try:
                currentContent = self.query_one(".content")
                currentContent.remove()
            except NoMatches:
                pass
            page_class = self.PAGES[activeIndex]["class"]
            page_instance = page_class(classes="content")
            self.mount(page_instance)

    def on_resize(self, event: events.Resize) -> None:
        console_size: Size = self.console.size
        aspect_ratio = (console_size.width / 2) / console_size.height
        if aspect_ratio < 1:
            self.layout = "v"
        else:
            self.layout = "h"

    # region callbacks
    # --------------- Callbacks ------------ #

    def action_goToTab(self, tab_number: int) -> None:
        """Go to the specified tab."""
        tabs = self.query_one(Tabs)
        tabs.active = f"t{tab_number}"

    def action_quit(self) -> None:
        self.exit()

    def action_go_to_categories(self) -> None:
        self.push_screen(CategoriesModal(), callback=self.on_categories_dismissed)

    def on_categories_dismissed(self, _) -> None:
        self.app.refresh(recompose=True)

    # region view
    # --------------- View --------------- #
    def compose(self) -> ComposeResult:
        with Container(classes="header"):
            yield Label(f"↪ {self.project_info['name']}", classes="title")
            yield Label(self.project_info["version"], classes="version")
            yield Label(get_user_host_string(), classes="user")
        yield Home(classes="content")
        yield Footer()



from datetime import datetime
from sqlalchemy import Column, DateTime, Integer, String
from sqlalchemy.orm import relationship
from .database.db import Base


class Person(Base):
    __tablename__ = "person"

    createdAt = Column(DateTime, nullable=False, default=datetime.now)
    updatedAt = Column(
        DateTime, nullable=False, default=datetime.now, onupdate=datetime.now
    )

    id = Column(Integer, primary_key=True, index=True)
    name = Column(String)

    splits = relationship("Split", back_populates="person")



from textual.app import ComposeResult
from textual.containers import Container, ScrollableContainer
from textual.widgets import (
    Footer,
)

from bagels.components.header import Header


class ModalContainer(ScrollableContainer, can_focus=False):
    # usage: ModalContainer(w1, w2, w3..... hotkeys=[])
    def __init__(self, *content, custom_classes: str = "wrapper max-width-60"):
        super().__init__(classes=custom_classes)
        self.content = content
        # self.hotkeys = hotkeys

    def compose(self) -> ComposeResult:
        yield Header(
            show_clock=False,
            icon="x",
            icon_action="app.pop_screen",
            icon_tooltip="Close",
        )
        with Container(classes="container"):
            for widget in self.content:
                yield widget
        yield Footer(show_command_palette=False)



from typing import Any, List, Union, Literal
from pydantic import BaseModel, Field
from rich.console import RenderableType


class Option(BaseModel):
    model_config = {"arbitrary_types_allowed": True}

    text: Union[str, None] = None  # if not provided, the value will be used
    value: Any
    prefix: Union[RenderableType, None] = None
    postfix: Union[RenderableType, None] = None


class Options(BaseModel):
    items: List[Option] = Field(default_factory=list)

    def __len__(self):
        return len(self.items)


class FormField(BaseModel):
    placeholder: Union[str, None] = None
    title: str
    key: str
    type: Literal[
        "string",
        "number",
        "integer",
        "boolean",
        "autocomplete",
        "dateAutoDay",
        "hidden",
    ]
    is_required: bool = False
    min: Union[float, int, None] = None
    max: Union[float, int, None] = None
    labels: Union[List[str], None] = None  # for type "boolean"
    options: Union[Options, None] = None  # for type "autocomplete"
    default_value: Any = None
    default_value_text: Union[str, None] = None
    create_action: Union[bool, None] = None  # for type "autocomplete"


class Form(BaseModel):
    fields: List[FormField] = Field(default_factory=list)

    def __len__(self):
        return len(self.fields)



from datetime import datetime
from sqlalchemy import (
    Column,
    DateTime,
    Integer,
    String,
    Float,
    Boolean,
    ForeignKey,
    CheckConstraint,
    event,
    func,
    select,
    update,
)
from sqlalchemy.orm import relationship, validates
from .database.db import Base


class RecordTemplate(Base):
    __tablename__ = "record_template"

    createdAt = Column(DateTime, nullable=False, default=datetime.now)
    updatedAt = Column(
        DateTime, nullable=False, default=datetime.now, onupdate=datetime.now
    )

    id = Column(Integer, primary_key=True, index=True)
    label = Column(String, nullable=False)
    amount = Column(Float, CheckConstraint("amount > 0"), nullable=False)
    accountId = Column(Integer, ForeignKey("account.id"), nullable=False)
    categoryId = Column(Integer, ForeignKey("category.id"), nullable=True)

    order = Column(Integer, nullable=False, unique=True)

    isIncome = Column(Boolean, nullable=False, default=False)
    isTransfer = Column(
        Boolean,
        CheckConstraint("(isTransfer = FALSE) OR (isIncome = FALSE)"),
        nullable=False,
        default=False,
    )
    transferToAccountId = Column(Integer, ForeignKey("account.id"), nullable=True)

    account = relationship("Account", foreign_keys=[accountId])
    category = relationship("Category", foreign_keys=[categoryId])
    transferToAccount = relationship("Account", foreign_keys=[transferToAccountId])

    def to_dict(self) -> dict:
        """Creates a dictionary object to feed into create_record."""
        return {
            "label": self.label,
            "amount": self.amount,
            "accountId": self.accountId,
            "categoryId": self.categoryId,
            "isIncome": self.isIncome,
            "isTransfer": self.isTransfer,
            "transferToAccountId": self.transferToAccountId,
        }

    @validates("order")
    def validate_order(self, key, order):
        if order is None:
            raise ValueError("Order cannot be null.")
        return order


@event.listens_for(RecordTemplate, "before_insert")
def receive_before_insert(mapper, connection, target):
    max_order = connection.execute(
        select(func.max(RecordTemplate.order))
    ).scalar_one_or_none()
    target.order = (max_order or 0) + 1
